"""Event base classes for Z-Wave JS."""


class EventBase:
    def __init__(self):
        """Initialize event base."""
        self._listeners = {}

    def on(self, event_name, callback):
        listeners: list = self._listeners.setdefault(event_name, [])
        listeners.append(callback)

        def unsub():
            if callback in listeners:
                listeners.remove(callback)

        return unsub

    def emit(self, event_name, data):
        for listener in self._listeners.get(event_name, []):
            listener(data)



"""Z-Wave JS Driver."""
from .controller import Controller
from ..event import EventBase


class Driver(EventBase):
    def __init__(self, state: dict):
        super().__init__()
        self.controller = Controller(state)

    def receive_event(self, event: dict):
        """Receive an event."""
        if event["source"] != "driver":
            self.controller.receive_event(event)
            return

        if event["event"] == "all nodes ready":
            pass

        self.emit(event["event"], event)



"""Basic CLI to test wrapper."""
import asyncio
import aiohttp
import logging
import sys

from .client import Client

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__package__)


async def main():
    if len(sys.argv) < 2:
        print("Error: pass URL to Z-Wave JS server")
        return

    async with aiohttp.ClientSession() as session:
        client = Client(sys.argv[1], session)
        asyncio.create_task(client.connect())

        setup = False

        def log_value_updated(event):
            node = event["node"]
            value = event["value"]

            if node.device_config:
                description = node.device_config["description"]
            else:
                description = f'{node.device_class["generic"]} (missing device config)'

            logger.info(
                "Node %s %s (%s) changed to %s",
                description,
                value.property_name or "",
                value.value_id,
                value.value,
            )

        while True:
            try:
                await asyncio.sleep(0.1)

                if not setup and client.driver:
                    setup = True

                    # Set up listeners on new nodes
                    client.driver.controller.on(
                        "node added",
                        lambda event: event["node"].on(
                            "value updated", log_value_updated
                        ),
                    )

                    # Set up listeners on existing nodes
                    for node in client.driver.controller.nodes.values():
                        node.on("value updated", log_value_updated)
            except KeyboardInterrupt:
                logger.info("Close requested")
                break

        await client.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



"""Client."""
import asyncio
import logging
import pprint
import random
from typing import Awaitable, Callable, List

from aiohttp import ClientSession, WSMsgType, client_exceptions

from .model.driver import Driver

STATE_CONNECTING = "connecting"
STATE_CONNECTED = "connected"
STATE_DISCONNECTED = "disconnected"


async def gather_callbacks(
    logger: logging.Logger, name: str, callbacks: List[Callable[[], Awaitable[None]]]
) -> None:
    results = await asyncio.gather(*[cb() for cb in callbacks], return_exceptions=True)
    for result, callback in zip(results, callbacks):
        if not isinstance(result, Exception):
            continue
        logger.error("Unexpected error in %s %s", name, callback, exc_info=result)


class NotConnected(Exception):
    """Exception raised when trying to handle unknown handler."""


class InvalidState(Exception):
    """Exception raised when data gets in invalid state."""


class Client:
    """Class to manage the IoT connection."""

    def __init__(self, ws_server_url: str, aiohttp_session: ClientSession):
        """Initialize the Client class."""
        self.ws_server_url = ws_server_url
        self.aiohttp_session = aiohttp_session
        self.driver = None
        # The WebSocket client
        self.client = None
        # Scheduled sleep task till next connection retry
        self.retry_task = None
        # Boolean to indicate if we wanted the connection to close
        self.close_requested = False
        # The current number of attempts to connect, impacts wait time
        self.tries = 0
        # Current state of the connection
        self.state = STATE_DISCONNECTED
        self._on_connect: List[Callable[[], Awaitable[None]]] = []
        self._on_disconnect: List[Callable[[], Awaitable[None]]] = []
        self._on_initialized: List[Callable[[], Awaitable[None]]] = []
        self._logger = logging.getLogger(__package__)
        self._disconnect_event = None

    def async_handle_message(self, msg) -> None:
        """Handle incoming message.
        Run all async tasks in a wrapper to log appropriately.
        """
        if msg["type"] == "state":
            if self.driver is None:
                self.driver = Driver(msg["state"])
                self._logger.info(
                    "Z-Wave JS initialized. %s nodes", len(self.driver.controller.nodes)
                )
                asyncio.create_task(
                    gather_callbacks(
                        self._logger, "on_initialized", self._on_initialized
                    )
                )
            else:
                # TODO how do we handle reconnect?
                pass

            return

        if self.driver is None:
            raise InvalidState("Did not receive state as first message")

        if msg["type"] != "event":
            # Can't handle
            return

        self.driver.receive_event(msg["event"])

    def register_on_connect(self, on_connect_cb: Callable[[], Awaitable[None]]):
        """Register an async on_connect callback."""
        self._on_connect.append(on_connect_cb)

    def register_on_disconnect(self, on_disconnect_cb: Callable[[], Awaitable[None]]):
        """Register an async on_disconnect callback."""
        self._on_disconnect.append(on_disconnect_cb)

    def register_on_initialized(self, on_initialized_cb: Callable[[], Awaitable[None]]):
        """Register an async on_initialized_cb callback."""
        self._on_initialized.append(on_initialized_cb)

    @property
    def connected(self):
        """Return if we're currently connected."""
        return self.state == STATE_CONNECTED

    async def async_send_json_message(self, message):
        """Send a message.

        Raises NotConnected if client not connected.
        """
        if self.state != STATE_CONNECTED:
            raise NotConnected

        if self._logger.isEnabledFor(logging.DEBUG):
            self._logger.debug("Publishing message:\n%s\n", pprint.pformat(message))

        await self.client.send_json(message)

    async def connect(self):
        """Connect to the IoT broker."""
        if self.state != STATE_DISCONNECTED:
            raise RuntimeError("Connect called while not disconnected")

        self.close_requested = False
        self.state = STATE_CONNECTING
        self.tries = 0
        self._disconnect_event = asyncio.Event()

        while True:
            try:
                self._logger.debug("Trying to connect")
                await self._handle_connection()
            except Exception:  # pylint: disable=broad-except
                # Safety net. This should never hit.
                # Still adding it here to make sure we can always reconnect
                self._logger.exception("Unexpected error")

            if self.state == STATE_CONNECTED and self._on_disconnect:
                await gather_callbacks(
                    self._logger, "on_disconnect", self._on_disconnect
                )

            if self.close_requested:
                break

            self.state = STATE_CONNECTING
            self.tries += 1

            try:
                await self._wait_retry()
            except asyncio.CancelledError:
                # Happens if disconnect called
                break

        self.state = STATE_DISCONNECTED
        self._disconnect_event.set()
        self._disconnect_event = None

    async def _wait_retry(self):
        """Wait until it's time till the next retry."""
        # Sleep 2^tries + 0…tries*3 seconds between retries
        self.retry_task = asyncio.create_task(
            asyncio.sleep(2 ** min(9, self.tries) + random.randint(0, self.tries * 3))
        )
        await self.retry_task
        self.retry_task = None

    async def _handle_connection(self):
        """Connect to the Z-Wave JS server."""
        client = None
        disconnect_warn = None
        try:
            self.client = client = await self.aiohttp_session.ws_connect(
                self.ws_server_url,
                heartbeat=55,
            )
            self.tries = 0

            self._logger.info("Connected")
            self.state = STATE_CONNECTED

            if self._on_connect:
                await gather_callbacks(self._logger, "on_connect", self._on_connect)

            while not client.closed:
                msg = await client.receive()

                if msg.type in (WSMsgType.CLOSED, WSMsgType.CLOSING):
                    break

                if msg.type == WSMsgType.ERROR:
                    disconnect_warn = "Connection error"
                    break

                if msg.type != WSMsgType.TEXT:
                    disconnect_warn = "Received non-Text message: {}".format(msg.type)
                    break

                try:
                    msg = msg.json()
                except ValueError:
                    disconnect_warn = "Received invalid JSON."
                    break

                if self._logger.isEnabledFor(logging.DEBUG):
                    self._logger.debug("Received message:\n%s\n", pprint.pformat(msg))

                try:
                    self.async_handle_message(msg)

                except InvalidState as err:
                    disconnect_warn = f"Invalid state: {err}"
                    await client.close()
                    break

                except Exception:  # pylint: disable=broad-except
                    self._logger.exception("Unexpected error handling %s", msg)
                    break

        except client_exceptions.WSServerHandshakeError as err:
            self._logger.warning("Unable to connect: %s", err)

        except client_exceptions.ClientError as err:
            self._logger.warning("Unable to connect: %s", err)

        except asyncio.CancelledError:
            pass

        finally:
            if disconnect_warn is None:
                self._logger.info("Connection closed")
            else:
                self._logger.warning("Connection closed: %s", disconnect_warn)

    async def disconnect(self):
        """Disconnect the client."""
        self.close_requested = True

        if self.client is not None:
            await self.client.close()
        elif self.retry_task is not None:
            self.retry_task.cancel()

        if self._disconnect_event is not None:
            await self._disconnect_event.wait()



"""Z-Wave node."""
from typing import List
from ..event import EventBase
from .value import Value, value_id


class Node(EventBase):
    def __init__(self, data: dict) -> None:
        """Initialize a node."""
        super().__init__()
        self.data = data
        self.values = {value_id(self, val): Value(self, val) for val in data["values"]}

    @property
    def node_id(self) -> int:
        """Return the node_id."""
        return self.data.get("nodeId")

    @property
    def index(self) -> int:
        """Return the index."""
        return self.data.get("index")

    @property
    def installer_icon(self) -> int:
        """Return the installer_icon."""
        return self.data.get("installerIcon")

    @property
    def user_icon(self) -> int:
        """Return the user_icon."""
        return self.data.get("userIcon")

    @property
    def status(self) -> int:
        """Return the status."""
        return self.data.get("status")

    @property
    def ready(self) -> bool:
        """Return the ready."""
        return self.data.get("ready")

    @property
    def device_class(self) -> dict:  # XXXXXX NEEDS CLASS
        """Return the device_class."""
        return self.data.get("deviceClass")

    @property
    def is_listening(self) -> bool:
        """Return the is_listening."""
        return self.data.get("isListening")

    @property
    def is_frequent_listening(self) -> bool:
        """Return the is_frequent_listening."""
        return self.data.get("isFrequentListening")

    @property
    def is_routing(self) -> bool:
        """Return the is_routing."""
        return self.data.get("isRouting")

    @property
    def max_baud_rate(self) -> int:
        """Return the max_baud_rate."""
        return self.data.get("maxBaudRate")

    @property
    def is_secure(self) -> bool:
        """Return the is_secure."""
        return self.data.get("isSecure")

    @property
    def version(self) -> int:
        """Return the version."""
        return self.data.get("version")

    @property
    def is_beaming(self) -> bool:
        """Return the is_beaming."""
        return self.data.get("isBeaming")

    @property
    def manufacturer_id(self) -> int:
        """Return the manufacturer_id."""
        return self.data.get("manufacturerId")

    @property
    def product_id(self) -> int:
        """Return the product_id."""
        return self.data.get("productId")

    @property
    def product_type(self) -> int:
        """Return the product_type."""
        return self.data.get("productType")

    @property
    def firmware_version(self) -> str:
        """Return the firmware_version."""
        return self.data.get("firmwareVersion")

    @property
    def zwave_plus_version(self) -> int:
        """Return the zwave_plus_version."""
        return self.data.get("zwavePlusVersion")

    @property
    def node_type(self) -> int:
        """Return the node_type."""
        return self.data.get("nodeType")

    @property
    def role_type(self) -> int:
        """Return the role_type."""
        return self.data.get("roleType")

    @property
    def name(self) -> str:
        """Return the name."""
        return self.data.get("name")

    @property
    def location(self) -> str:
        """Return the location."""
        return self.data.get("location")

    @property
    def device_config(self) -> dict:  # XXXXXX NEEDS CLASS
        """Return the device_config."""
        return self.data.get("deviceConfig")

    @property
    def label(self) -> str:
        """Return the label."""
        return self.data.get("label")

    @property
    def neighbors(self) -> List[int]:
        """Return the neighbors."""
        return self.data.get("neighbors")

    @property
    def endpoint_count_is_dynamic(self) -> bool:
        """Return the endpoint_count_is_dynamic."""
        return self.data.get("endpointCountIsDynamic")

    @property
    def endpoints_have_identical_capabilities(self) -> bool:
        """Return the endpoints_have_identical_capabilities."""
        return self.data.get('"endpointsHaveIdenticalCapabilities')

    @property
    def individual_endpoint_count(self) -> int:
        """Return the individual_endpoint_count."""
        return self.data.get("individualEndpointCount")

    @property
    def aggregated_endpoint_count(self) -> int:
        """Return the aggregated_endpoint_count."""
        return self.data.get("aggregatedEndpointCount")

    @property
    def interview_attempts(self) -> int:
        """Return the interview_attempts."""
        return self.data.get("interviewAttempts")

    def receive_event(self, event: dict):
        """Receive an event."""
        if event["event"] == "value added":
            value = Value(self, event["args"])
            self.values[value.value_id] = event["value"] = value

        elif event["event"] == "value updated":
            value = self.values.get(value_id(self, event["args"]))
            if value is None:
                # TODO decide how to handle value updated for unknown values
                print()
                print("Value updated for unknown value", value_id(self, event["args"]))
                print("Available value IDs", ", ".join(self.values))
                print()
                value = Value(self, event["args"])
                self.values[value.value_id] = value
            else:
                value.receive_event(event)
            event["value"] = value

        elif event["event"] == "value removed":
            event["value"] = self.values.pop(value_id(self, event["args"]))

        elif event["event"] == "metadata updated":
            pass

        elif event["event"] == "value notification":
            pass

        elif event["event"] == "notification":
            pass

        elif event["event"] == "interview failed":
            pass

        elif event["event"] == "firmware update progress":
            pass

        elif event["event"] == "firmware update finished":
            pass

        elif event["event"] == "wake up":
            pass

        elif event["event"] == "sleep":
            pass

        elif event["event"] == "dead":
            pass

        elif event["event"] == "alive":
            pass

        elif event["event"] == "interview completed":
            pass

        elif event["event"] == "ready":
            pass

        else:
            # TODO decide what to do with unknown event
            print(f"Unhandled node event for node {self.node_id}: {event}")

        event["node"] = self

        self.emit(event["event"], event)



"""Value model."""


from typing import Any, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .node import Node


def value_id(node: "Node", event: dict) -> str:
    """Return ID of value."""
    command_class = event.get("commandClass")
    endpoint = event.get("endpoint")
    property_key_name = event.get("propertyKeyName")
    return f"{node.node_id}-{command_class}-{endpoint}-{property_key_name}"


class ValueMetadata:
    def __init__(self, data) -> None:
        self.data = data

    @property
    def type(self) -> str:
        """Return type."""
        return self.data.get("type")

    @property
    def readable(self) -> bool:
        """Return readable."""
        return self.data.get("readable")

    @property
    def writeable(self) -> bool:
        """Return writeable."""
        return self.data.get("writeable")

    @property
    def label(self) -> str:
        """Return label."""
        return self.data.get("label")

    @property
    def min(self) -> Optional[int]:
        """Return min."""
        return self.data.get("min")

    @property
    def max(self) -> Optional[int]:
        """Return max."""
        return self.data.get("max")

    @property
    def unit(self) -> Optional[str]:
        """Return unit."""
        return self.data.get("unit")

    @property
    def cc_specific(self) -> Optional[dict]:
        """Return ccSpecific."""
        return self.data.get("ccSpecific")


class Value:
    def __init__(self, node: "Node", data: dict) -> None:
        self.node = node
        self.data = data
        self._value = data.get("value")

    @property
    def value_id(self) -> str:
        """Return value ID."""
        return value_id(self.node, self.data)

    @property
    def metadata(self) -> ValueMetadata:
        """Return value metadata."""
        return ValueMetadata(self.data["metadata"])

    @property
    def value(self) -> Optional[Any]:
        """Return value."""
        return self._value

    @property
    def command_class_name(self) -> str:
        """Return commandClassName."""
        return self.data.get("commandClassName")

    @property
    def command_class(self) -> int:
        """Return commandClass."""
        return self.data.get("commandClass")

    @property
    def endpoint(self) -> int:
        """Return endpoint."""
        return self.data.get("endpoint")

    @property
    def property_(self) -> str:
        """Return property."""
        return self.data.get("property")

    @property
    def property_key(self) -> str:
        """Return propertyKey."""
        return self.data.get("propertyKey")

    @property
    def property_name(self) -> str:
        """Return propertyName."""
        return self.data.get("propertyName")

    @property
    def property_key_name(self) -> str:
        """Return propertyKeyName."""
        return self.data.get("propertyKeyName")

    def receive_event(self, event: dict):
        """Receive an event."""
        if event["event"] != "value updated":
            return

        self.data.update(event["args"])
        self._value = event["args"].get("newValue")



from typing import Dict, List
from ..event import EventBase
from .node import Node


class Controller(EventBase):
    def __init__(self, state: dict) -> None:
        """Initialize controller."""
        super().__init__()
        self.data = state["controller"]
        self.nodes: Dict[int, Node] = {}
        for node_state in state["nodes"]:
            node = Node(node_state)
            self.nodes[node.node_id] = node

    @property
    def library_version(self) -> str:
        """Return library_version."""
        return self.data.get("libraryVersion")

    @property
    def controller_type(self) -> int:
        """Return controller_type."""
        return self.data.get("type")

    @property
    def home_id(self) -> int:
        """Return home_id."""
        return self.data.get("homeId")

    @property
    def own_node_id(self) -> int:
        """Return own_node_id."""
        return self.data.get("ownNodeId")

    @property
    def is_secondary(self) -> bool:
        """Return is_secondary."""
        return self.data.get("isSecondary")

    @property
    def is_using_home_id_from_other_network(self) -> bool:
        """Return is_using_home_id_from_other_network."""
        return self.data.get("isUsingHomeIdFromOtherNetwork")

    @property
    def is_SIS_present(self) -> bool:
        """Return is_SIS_present."""
        return self.data.get("isSISPresent")

    @property
    def was_real_primary(self) -> bool:
        """Return was_real_primary."""
        return self.data.get("wasRealPrimary")

    @property
    def is_static_update_controller(self) -> bool:
        """Return is_static_update_controller."""
        return self.data.get("isStaticUpdateController")

    @property
    def is_slave(self) -> bool:
        """Return is_slave."""
        return self.data.get("isSlave")

    @property
    def serial_api_version(self) -> str:
        """Return serial_api_version."""
        return self.data.get("serialApiVersion")

    @property
    def manufacturer_id(self) -> int:
        """Return manufacturer_id."""
        return self.data.get("manufacturerId")

    @property
    def product_type(self) -> int:
        """Return product_type."""
        return self.data.get("productType")

    @property
    def product_id(self) -> int:
        """Return product_id."""
        return self.data.get("productId")

    @property
    def supported_function_types(self) -> List[int]:
        """Return supported_function_types."""
        return self.data.get("supportedFunctionTypes")

    @property
    def suc_node_id(self) -> int:
        """Return suc_node_id."""
        return self.data.get("sucNodeId")

    @property
    def supports_timers(self) -> bool:
        """Return supports_timers."""
        return self.data.get("supportsTimers")

    def receive_event(self, event: dict):
        """Receive an event."""
        if event["source"] == "node":
            node = self.nodes.get(event["nodeId"])
            if node is None:
                # TODO handle event for uknown node
                pass
            else:
                node.receive_event(event)
            return

        if event["source"] != "controller":
            # TODO decide what to do here
            print(f"Controller doesn't know how to handle/forward this event: {event}")

        if event["event"] == "inclusion failed":
            pass

        elif event["event"] == "exclusion failed":
            pass

        elif event["event"] == "inclusion started":
            pass

        elif event["event"] == "exclusion started":
            pass

        elif event["event"] == "inclusion stopped":
            pass

        elif event["event"] == "exclusion stopped":
            pass

        elif event["event"] == "node added":
            node = event["node"] = Node(event["node"])
            self.nodes[node.node_id] = node

        elif event["event"] == "node removed":
            event["node"] = self.nodes.pop(event["node"]["nodeId"])

        elif event["event"] == "heal network progress":
            pass

        elif event["event"] == "heal network done":
            pass

        else:
            # TODO decide what to do with unknown event
            print(f"Unhandled node event for controller: {event}")

        event["controller"] = self

        self.emit(event["event"], event)
